"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from types import SimpleNamespace
from typing import Any
from unittest.mock import MagicMock, patch

import pytest
from microsoft.teams.apps.routing.activity_context import ActivityContext


class TestOptionalGraphDependencies:
    """Test that graph functionality is properly optional."""

    def _create_activity_context(self) -> ActivityContext[Any]:
        """Create a minimal ActivityContext for testing."""
        # Create mock objects for all required parameters
        mock_activity = MagicMock()
        mock_logger = MagicMock()
        mock_storage = MagicMock()
        mock_api = MagicMock()
        mock_conversation_ref = MagicMock()
        mock_sender = MagicMock()
        mock_app_token = MagicMock()  # Provide an app token for graph access

        return ActivityContext(
            activity=mock_activity,
            app_id="test-app-id",
            logger=mock_logger,
            storage=mock_storage,
            api=mock_api,
            user_token=None,
            conversation_ref=mock_conversation_ref,
            is_signed_in=False,
            connection_name="test-connection",
            sender=mock_sender,
            app_token=mock_app_token,  # This is needed for app_graph to work
        )

    def test_app_graph_property_without_graph_available(self) -> None:
        """Test app_graph property when graph dependencies are not available."""

        # Mock import error for graph module
        def mock_import(name: str, *args: Any, **kwargs: Any) -> Any:
            if name == "microsoft.teams.graph":
                raise ImportError("No module named 'microsoft.teams.graph'")
            return __import__(name, *args, **kwargs)

        with patch("builtins.__import__", side_effect=mock_import):
            activity_context = self._create_activity_context()
            # app_graph should raise RuntimeError when graph dependencies are not available
            with pytest.raises(RuntimeError, match="Failed to create app graph client"):
                _ = activity_context.app_graph

    def test_app_graph_property_with_graph_available(self) -> None:
        """Test app_graph property when graph dependencies are available."""

        # Mock successful graph module import
        def mock_import(name: str, *args: Any, **kwargs: Any) -> Any:
            if name == "microsoft.teams.graph":
                # Create a mock module with get_graph_client
                mock_module = SimpleNamespace()
                mock_module.get_graph_client = lambda x: "MockGraphClient"  # type: ignore
                return mock_module
            return __import__(name, *args, **kwargs)

        with patch("builtins.__import__", side_effect=mock_import):
            activity_context = self._create_activity_context()
            result = activity_context.app_graph
            assert result == "MockGraphClient"

    def test_user_graph_property_not_signed_in(self) -> None:
        """Test user_graph property when user is not signed in."""
        activity_context = ActivityContext(
            activity=MagicMock(),
            app_id="test-app-id",
            logger=MagicMock(),
            storage=MagicMock(),
            api=MagicMock(),
            user_token=MagicMock(),  # Has token but not signed in
            conversation_ref=MagicMock(),
            is_signed_in=False,  # Not signed in
            connection_name="test-connection",
            sender=MagicMock(),
            app_token=None,
        )

        # user_graph should raise ValueError when user is not signed in
        with pytest.raises(ValueError, match="User must be signed in to access Graph client"):
            _ = activity_context.user_graph

    def test_user_graph_property_no_token(self) -> None:
        """Test user_graph property when user is signed in but has no token."""
        activity_context = ActivityContext(
            activity=MagicMock(),
            app_id="test-app-id",
            logger=MagicMock(),
            storage=MagicMock(),
            api=MagicMock(),
            user_token=None,  # No token
            conversation_ref=MagicMock(),
            is_signed_in=True,  # Signed in but no token
            connection_name="test-connection",
            sender=MagicMock(),
            app_token=None,
        )

        # user_graph should raise ValueError when no user token is available
        with pytest.raises(ValueError, match="No user token available for Graph client"):
            _ = activity_context.user_graph

    def test_app_graph_property_no_token(self) -> None:
        """Test app_graph property when no app token is available."""
        activity_context = ActivityContext(
            activity=MagicMock(),
            app_id="test-app-id",
            logger=MagicMock(),
            storage=MagicMock(),
            api=MagicMock(),
            user_token=None,
            conversation_ref=MagicMock(),
            is_signed_in=False,
            connection_name="test-connection",
            sender=MagicMock(),
            app_token=None,  # No app token
        )

        # app_graph should raise ValueError when no app token is available
        with pytest.raises(RuntimeError, match="Token cannot be None"):
            _ = activity_context.app_graph



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

import asyncio
import random
from logging import Logger
from typing import Awaitable, Callable, Literal, Optional, TypeVar

from microsoft.teams.common import ConsoleLogger

T = TypeVar("T")

# Type alias for jitter types using string literals
JitterType = Literal["none", "full", "equal", "decorrelated"]


class RetryOptions:
    def __init__(
        self,
        max_attempts: int = 5,
        delay: float = 0.5,  # in seconds
        max_delay: float = 30.0,  # maximum delay cap
        jitter_type: JitterType = "full",
        logger: Optional[Logger] = None,
        previous_delay: Optional[float] = None,  # Internal use for decorrelated jitter
        attempt_number: int = 1,  # Internal use to track current attempt number
        _internal_logger: Optional[Logger] = None,  # Internal use to pass existing child logger
    ):
        self.max_attempts = max_attempts
        self.delay = delay
        self.max_delay = max_delay
        self.jitter_type: JitterType = jitter_type
        # Use existing internal logger if provided, otherwise create child logger
        self.logger = (
            _internal_logger
            if _internal_logger
            else (logger.getChild("@teams/retry") if logger else ConsoleLogger().create_logger("@teams/retry"))
        )
        self.previous_delay = previous_delay
        self.attempt_number = attempt_number


def _apply_jitter(delay: float, jitter_type: JitterType, previous_delay: Optional[float] = None) -> float:
    """Apply jitter to the delay to prevent thundering herd problems."""
    if jitter_type == "none":
        return delay
    elif jitter_type == "full":
        # Random delay between 0 and the computed delay
        return random.uniform(0, delay)
    elif jitter_type == "equal":
        # Random delay between half the computed delay and the full computed delay
        return random.uniform(delay / 2, delay)
    elif jitter_type == "decorrelated":
        # Decorrelated jitter: random delay between base delay and 3 * previous delay
        base_delay = delay / 4  # Start with a smaller base
        if previous_delay is None:
            previous_delay = base_delay
        return random.uniform(base_delay, min(delay, 3 * previous_delay))
    else:
        return delay


async def retry(factory: Callable[[], Awaitable[T]], options: Optional[RetryOptions] = None) -> T:
    options = options or RetryOptions()
    max_attempts = options.max_attempts
    base_delay = options.delay
    max_delay = options.max_delay
    jitter_type: JitterType = options.jitter_type
    logger = options.logger
    previous_delay = options.previous_delay
    attempt_number = options.attempt_number

    try:
        return await factory()
    except (asyncio.CancelledError, KeyboardInterrupt):
        # Don't retry cancellation or keyboard interrupts
        logger.debug("Operation cancelled or interrupted, not retrying")
        raise
    except Exception as err:
        if max_attempts > 1:
            # Calculate exponential backoff delay using the attempt number
            exponential_delay = base_delay * (2 ** (attempt_number - 1))

            # Cap the delay at max_delay
            capped_delay = min(exponential_delay, max_delay)

            # Apply jitter
            jittered_delay = _apply_jitter(capped_delay, jitter_type, previous_delay)

            logger.debug(f"Delaying {jittered_delay:.2f}s before retry (attempt {attempt_number})...")
            await asyncio.sleep(jittered_delay)
            logger.debug("Retrying...")

            return await retry(
                factory,
                RetryOptions(
                    max_attempts=max_attempts - 1,
                    delay=base_delay,
                    max_delay=max_delay,
                    jitter_type=jitter_type,
                    previous_delay=jittered_delay,  # Pass current delay for decorrelated jitter
                    attempt_number=attempt_number + 1,  # Increment attempt number
                    _internal_logger=logger,  # Pass the existing logger to avoid nested children
                ),
            )
        logger.error("Final attempt failed.", exc_info=err)
        raise



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from dataclasses import dataclass, field
from logging import Logger
from typing import Any, Awaitable, Callable, List, Optional, TypedDict, Union, cast

from microsoft.teams.common import Storage
from typing_extensions import Unpack

from .plugins import PluginBase


class AppOptions(TypedDict, total=False):
    """Configuration options for the Teams App."""

    # Authentication credentials
    client_id: Optional[str]
    client_secret: Optional[str]
    tenant_id: Optional[str]
    # Custom token provider function
    token: Optional[Callable[[Union[str, list[str]], Optional[str]], Union[str, Awaitable[str]]]]

    # Infrastructure
    logger: Optional[Logger]
    storage: Optional[Storage[str, Any]]
    plugins: Optional[List[PluginBase]]
    skip_auth: Optional[bool]

    # Oauth
    default_connection_name: Optional[str]


@dataclass
class InternalAppOptions:
    """Internal dataclass for AppOptions with defaults and non-nullable fields."""

    # Fields with defaults
    skip_auth: bool = False
    default_connection_name: str = "graph"
    plugins: List[PluginBase] = field(default_factory=lambda: [])

    # Optional fields
    client_id: Optional[str] = None
    client_secret: Optional[str] = None
    tenant_id: Optional[str] = None
    token: Optional[Callable[[Union[str, list[str]], Optional[str]], Union[str, Awaitable[str]]]] = None
    logger: Optional[Logger] = None
    storage: Optional[Storage[str, Any]] = None

    @classmethod
    def from_typeddict(cls, options: AppOptions) -> "InternalAppOptions":
        """
        Create InternalAppOptions from AppOptions TypedDict with defaults applied.

        Args:
            options: AppOptions TypedDict (potentially with None values)

        Returns:
            InternalAppOptions with proper defaults and non-nullable required fields
        """
        kwargs: dict[str, Any] = {k: v for k, v in options.items() if v is not None}
        return cls(**kwargs)


def merge_app_options_with_defaults(**options: Unpack[AppOptions]) -> AppOptions:
    """
    Create AppOptions with default values merged with provided options.

    Args:
        **options: Configuration options to override defaults

    Returns:
        AppOptions with defaults applied
    """
    defaults: AppOptions = {
        "skip_auth": False,
        "default_connection_name": "graph",
        "plugins": [],
    }

    return cast(AppOptions, {**defaults, **options})



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from microsoft.teams.api import ClientCredentials, JsonWebToken
from microsoft.teams.apps.token_manager import TokenManager
from microsoft.teams.common import Client

# Valid JWT-like token for testing (format: header.payload.signature)
VALID_TEST_TOKEN = (
    "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9."
    "eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ."
    "SflKxwRJSMeKKF2QT4fwpMeJf36POk6yJV_adQssw5c"
)


class TestTokenManager:
    """Test TokenManager functionality."""

    @pytest.mark.asyncio
    async def test_get_bot_token_success(self):
        """Test successful bot token refresh, caching, and expiration refresh."""
        # First token response
        mock_token_response1 = MagicMock()
        mock_token_response1.access_token = VALID_TEST_TOKEN

        # Second token response for expired token
        mock_token_response2 = MagicMock()
        mock_token_response2.access_token = (
            "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9."
            "eyJzdWIiOiI5ODc2NTQzMjEwIiwibmFtZSI6IkphbmUgRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ."
            "Twzj7LKlhYUUe2GFRME4WOZdWq2TdayZhWjhBr1r5X4"
        )

        # Third token response for force refresh
        mock_token_response3 = MagicMock()
        mock_token_response3.access_token = (
            "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9."
            "eyJzdWIiOiIxMTExMTExMTExIiwibmFtZSI6IkZvcmNlIFJlZnJlc2giLCJpYXQiOjE1MTYyMzkwMjJ9."
            "dQw4w9WgXcQ"
        )

        mock_credentials = ClientCredentials(
            client_id="test-client-id",
            client_secret="test-client-secret",
            tenant_id="test-tenant-id",
        )

        # Mock the BotTokenClient
        mock_bot_token_client = MagicMock()
        mock_bot_token_client.get = AsyncMock(
            side_effect=[mock_token_response1, mock_token_response2, mock_token_response3]
        )

        mock_http_client = MagicMock(spec=Client)
        mock_http_client.clone = MagicMock(return_value=mock_http_client)

        with patch("microsoft.teams.apps.token_manager.BotTokenClient", return_value=mock_bot_token_client):
            manager = TokenManager(
                http_client=mock_http_client,
                credentials=mock_credentials,
            )

            # First call
            token1 = await manager.get_bot_token()
            assert token1 is not None
            assert isinstance(token1, JsonWebToken)
            mock_bot_token_client.get.assert_called_once()

            # Second call should use cache (mock should still only be called once)
            token2 = await manager.get_bot_token()
            assert token2 == token1
            mock_bot_token_client.get.assert_called_once()  # Still only called once due to caching

            # Mock the token as expired
            token1.is_expired = MagicMock(return_value=True)

            # Third call should refresh because token is expired
            token3 = await manager.get_bot_token()
            assert token3 is not None
            assert token3 != token1  # New token
            assert mock_bot_token_client.get.call_count == 2

            # Force refresh even if not expired
            token3.is_expired = MagicMock(return_value=False)
            token4 = await manager.get_bot_token(force=True)
            assert token4 is not None
            assert mock_bot_token_client.get.call_count == 3

    @pytest.mark.asyncio
    async def test_get_bot_token_no_credentials(self):
        """Test refreshing bot token with no credentials returns None."""
        mock_http_client = MagicMock(spec=Client)
        mock_http_client.clone = MagicMock(return_value=mock_http_client)

        with patch("microsoft.teams.apps.token_manager.BotTokenClient"):
            manager = TokenManager(
                http_client=mock_http_client,
                credentials=None,
            )

            token = await manager.get_bot_token()
            assert token is None

    @pytest.mark.asyncio
    async def test_get_graph_token_default(self):
        """Test getting default graph token with caching and expiration refresh."""
        # First token response
        mock_token_response1 = MagicMock()
        mock_token_response1.access_token = VALID_TEST_TOKEN

        # Second token response for expired token
        mock_token_response2 = MagicMock()
        mock_token_response2.access_token = (
            "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9."
            "eyJzdWIiOiI5ODc2NTQzMjEwIiwibmFtZSI6IkphbmUgRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ."
            "Twzj7LKlhYUUe2GFRME4WOZdWq2TdayZhWjhBr1r5X4"
        )

        mock_credentials = ClientCredentials(
            client_id="test-client-id",
            client_secret="test-client-secret",
            tenant_id="default-tenant-id",
        )

        # Mock the BotTokenClient
        mock_bot_token_client = MagicMock()
        mock_bot_token_client.get_graph = AsyncMock(side_effect=[mock_token_response1, mock_token_response2])

        mock_http_client = MagicMock(spec=Client)
        mock_http_client.clone = MagicMock(return_value=mock_http_client)

        with patch("microsoft.teams.apps.token_manager.BotTokenClient", return_value=mock_bot_token_client):
            manager = TokenManager(
                http_client=mock_http_client,
                credentials=mock_credentials,
            )

            token1 = await manager.get_graph_token()

            assert token1 is not None
            assert isinstance(token1, JsonWebToken)

            # Verify it's cached
            token2 = await manager.get_graph_token()
            assert token2 == token1
            mock_bot_token_client.get_graph.assert_called_once()

            # Mock the token as expired
            token1.is_expired = MagicMock(return_value=True)

            # Third call should refresh because token is expired
            token3 = await manager.get_graph_token()
            assert token3 is not None
            assert token3 != token1  # New token
            assert mock_bot_token_client.get_graph.call_count == 2

    @pytest.mark.asyncio
    async def test_get_graph_token_with_tenant(self):
        """Test getting tenant-specific graph token."""
        mock_token_response = MagicMock()
        mock_token_response.access_token = VALID_TEST_TOKEN

        original_credentials = ClientCredentials(
            client_id="test-client-id",
            client_secret="test-client-secret",
            tenant_id="original-tenant-id",
        )

        # Mock the BotTokenClient
        mock_bot_token_client = MagicMock()
        mock_bot_token_client.get_graph = AsyncMock(return_value=mock_token_response)

        mock_http_client = MagicMock(spec=Client)
        mock_http_client.clone = MagicMock(return_value=mock_http_client)

        with patch("microsoft.teams.apps.token_manager.BotTokenClient", return_value=mock_bot_token_client):
            manager = TokenManager(
                http_client=mock_http_client,
                credentials=original_credentials,
            )

            token = await manager.get_graph_token("different-tenant-id")

            assert token is not None
            mock_bot_token_client.get_graph.assert_called_once()

            # Verify tenant-specific credentials were created
            call_args = mock_bot_token_client.get_graph.call_args
            passed_credentials = call_args[0][0]
            assert isinstance(passed_credentials, ClientCredentials)
            assert passed_credentials.tenant_id == "different-tenant-id"

    @pytest.mark.asyncio
    async def test_graph_token_force_refresh(self):
        """Test force refreshing graph token even when not expired."""
        mock_token_response1 = MagicMock()
        mock_token_response1.access_token = VALID_TEST_TOKEN

        mock_token_response2 = MagicMock()
        mock_token_response2.access_token = (
            "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9."
            "eyJzdWIiOiIxMTExMTExMTExIiwibmFtZSI6IkZvcmNlIFJlZnJlc2giLCJpYXQiOjE1MTYyMzkwMjJ9."
            "dQw4w9WgXcQ"
        )

        mock_credentials = ClientCredentials(
            client_id="test-client-id",
            client_secret="test-client-secret",
            tenant_id="test-tenant-id",
        )

        mock_bot_token_client = MagicMock()
        mock_bot_token_client.get_graph = AsyncMock(side_effect=[mock_token_response1, mock_token_response2])

        mock_http_client = MagicMock(spec=Client)
        mock_http_client.clone = MagicMock(return_value=mock_http_client)

        with patch("microsoft.teams.apps.token_manager.BotTokenClient", return_value=mock_bot_token_client):
            manager = TokenManager(
                http_client=mock_http_client,
                credentials=mock_credentials,
            )

            # First call
            token1 = await manager.get_graph_token()
            assert token1 is not None
            mock_bot_token_client.get_graph.assert_called_once()

            # Second call should use cache
            token2 = await manager.get_graph_token()
            assert token2 == token1
            mock_bot_token_client.get_graph.assert_called_once()  # Still only called once

            # Force refresh should call API even if not expired
            token3 = await manager.get_graph_token(force=True)
            assert token3 is not None
            assert mock_bot_token_client.get_graph.call_count == 2  # Now called twice



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import Any
from unittest.mock import AsyncMock, MagicMock

import pytest
from microsoft.teams.api import ApiClient
from microsoft.teams.api.activities.message.message import MessageActivityInput
from microsoft.teams.api.models.conversation.conversation_reference import ConversationReference
from microsoft.teams.apps import FunctionContext
from microsoft.teams.cards import AdaptiveCard


@pytest.mark.asyncio
class TestFunctionContextSend:
    """Test cases for FunctionContext."""

    @pytest.fixture
    def mock_api(self):
        """Create a mock ApiClient."""
        api = MagicMock(spec=ApiClient)
        api.service_url = "https://test.service.url"
        mock_conversations = MagicMock()
        mock_conversations.create = AsyncMock(return_value=MagicMock(id="new-conv"))
        mock_conversations.members_client.get_by_id = AsyncMock(return_value=True)

        api.conversations = mock_conversations
        return api

    @pytest.fixture
    def mock_http(self):
        """Create a mock HttpPlugin."""
        http = MagicMock()
        http.send = AsyncMock()
        http.send.return_value = "sent-activity"
        return http

    @pytest.fixture
    def mock_logger(self):
        """Create a mock Logger."""
        return MagicMock()

    @pytest.fixture
    def function_context(self, mock_api: ApiClient, mock_http: Any, mock_logger: Any) -> FunctionContext[Any]:
        ctx: FunctionContext[Any] = FunctionContext(
            id="bot-123",
            name="Test Bot",
            api=mock_api,
            http=mock_http,
            log=mock_logger,
            data={"some": "payload"},
            app_session_id="dummy-session",
            tenant_id="tenant-789",
            user_id="user-456",
            user_name="Test User",
            page_id="page-001",
            auth_token="token-abc",
            chat_id="conv-123",
        )
        return ctx

    async def test_send_string_activity(
        self,
        function_context: FunctionContext[Any],
        mock_http: Any,
    ) -> None:
        """Test sending a string message."""

        result = await function_context.send("Hello world")
        assert result == "sent-activity"

        sent_activity, conversation_ref = mock_http.send.call_args[0]

        assert isinstance(sent_activity, MessageActivityInput)
        assert sent_activity.text == "Hello world"

        assert isinstance(conversation_ref, ConversationReference)
        assert conversation_ref.conversation.id == "conv-123"

    async def test_send_adaptive_card(
        self,
        function_context: FunctionContext[Any],
        mock_http: Any,
    ) -> None:
        """Test sending an AdaptiveCard message."""

        card = AdaptiveCard(schema="1.0")
        result = await function_context.send(card)
        assert result == "sent-activity"

        sent_activity, conversation_ref = mock_http.send.call_args[0]

        assert sent_activity.attachments[0].content == card
        assert conversation_ref.conversation.id == "conv-123"

    async def test_send_creates_conversation_if_none(
        self, function_context: FunctionContext[Any], mock_http: Any
    ) -> None:
        """Test send() creates a conversation when _resolved_conversation_id is None."""

        function_context.chat_id = None

        mock_http.send.return_value = "sent-new-conv"

        result = await function_context.send("Hello new conversation")

        assert result == "sent-new-conv"
        # Ensure conversation was created
        assert function_context.api.conversations.create.call_count == 1  # type: ignore
        sent_activity, conversation_ref = mock_http.send.call_args[0]
        assert sent_activity.text == "Hello new conversation"
        assert conversation_ref.conversation.id == "new-conv"



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

import logging
from typing import Optional

from microsoft.teams.api import (
    BotTokenClient,
    ClientCredentials,
    Credentials,
    JsonWebToken,
    TokenProtocol,
)
from microsoft.teams.common import Client, ConsoleLogger, LocalStorage, LocalStorageOptions


class TokenManager:
    """Manages authentication tokens for the Teams application."""

    def __init__(
        self,
        http_client: Client,
        credentials: Optional[Credentials],
        logger: Optional[logging.Logger] = None,
        default_connection_name: Optional[str] = None,
    ):
        self._bot_token_client = BotTokenClient(http_client.clone())
        self._credentials = credentials
        self._default_connection_name = default_connection_name

        if not logger:
            self._logger = ConsoleLogger().create_logger("TokenManager")
        else:
            self._logger = logger.getChild("TokenManager")

        self._bot_token: Optional[TokenProtocol] = None

        # Key: tenant_id (empty string "" for default app graph token)
        self._graph_tokens: LocalStorage[TokenProtocol] = LocalStorage({}, LocalStorageOptions(max=20000))

    async def get_bot_token(self, force: bool = False) -> Optional[TokenProtocol]:
        """Refresh the bot authentication token."""
        if not self._credentials:
            self._logger.warning("No credentials provided, skipping bot token refresh")
            return None

        if not force and self._bot_token and not self._bot_token.is_expired():
            return self._bot_token

        if self._bot_token:
            self._logger.debug("Refreshing bot token")
        else:
            self._logger.debug("Retrieving bot token")

        token_response = await self._bot_token_client.get(self._credentials)
        self._bot_token = JsonWebToken(token_response.access_token)
        self._logger.debug("Bot token refreshed successfully")
        return self._bot_token

    async def get_graph_token(self, tenant_id: Optional[str] = None, force: bool = False) -> Optional[TokenProtocol]:
        """
        Get or refresh a Graph API token.

        Args:
            tenant_id: If provided, gets a tenant-specific token. Otherwise uses app's default.
            force: Force refresh even if token is not expired

        Returns:
            The graph token or None if not available
        """
        if not self._credentials:
            self._logger.debug("No credentials provided for graph token refresh")
            return None

        # Use empty string as key for default graph token
        key = tenant_id or ""

        cached = self._graph_tokens.get(key)
        if not force and cached and not cached.is_expired():
            return cached

        creds = self._credentials
        if tenant_id and isinstance(self._credentials, ClientCredentials):
            creds = ClientCredentials(
                client_id=self._credentials.client_id,
                client_secret=self._credentials.client_secret,
                tenant_id=tenant_id,
            )

        response = await self._bot_token_client.get_graph(creds)
        token = JsonWebToken(response.access_token)
        self._graph_tokens.set(key, token)

        self._logger.debug(f"Refreshed graph token tenant_id={tenant_id}")

        return token



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import Awaitable, Callable, Literal, Optional, Protocol, Union

from microsoft.teams.api import MessageActivityInput, SentActivity, TypingActivityInput

StreamerEvent = Literal["chunk", "close"]


class StreamerProtocol(Protocol):
    """Component that can send streamed chunks of an activity."""

    @property
    def closed(self) -> bool:
        """Whether the final stream message has been sent."""
        ...

    @property
    def count(self) -> int:
        """The total number of chunks queued to be sent."""
        ...

    @property
    def sequence(self) -> int:
        """
        The sequence number, representing the number of stream activities sent.

        Several chunks can be aggregated into one stream activity
        due to differences in Api rate limits.
        """
        ...

    def on_chunk(self, handler: Callable[[SentActivity], Awaitable[None]]) -> None:
        """
        Register a handler for chunk events.

        Args:
            handler: Async function that will be called for each chunk activity
        """
        ...

    def on_close(self, handler: Callable[[SentActivity], Awaitable[None]]) -> None:
        """
        Register a handler for close events.

        Args:
            handler: Async function that will be called when the stream closes
        """
        ...

    def emit(self, activity: Union[MessageActivityInput, TypingActivityInput, str]) -> None:
        """
        Emit an activity chunk.
        """
        ...

    def update(self, text: str) -> None:
        """
        Send status updates before emitting (ex. "Thinking...").

        Args:
            text: The status text to send.
        """
        ...

    async def close(self) -> Optional[SentActivity]:
        """
        Close the stream.
        """
        ...



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest.mock import MagicMock

import pytest
from microsoft.teams.apps.utils import RetryOptions, retry

# pyright: basic


class TestRetryOptions:
    """Test RetryOptions functionality."""

    def test_default_initialization(self):
        """Test that RetryOptions initializes with correct defaults."""
        options = RetryOptions()
        assert options.max_attempts == 5
        assert options.delay == 0.5
        assert options.max_delay == 30.0
        assert options.jitter_type == "full"
        assert options.attempt_number == 1
        assert options.logger is not None

    def test_custom_initialization(self):
        """Test that RetryOptions can be initialized with custom values."""
        mock_logger = MagicMock()
        options = RetryOptions(
            max_attempts=3,
            delay=1.0,
            max_delay=60.0,
            jitter_type="none",
            logger=mock_logger,
            attempt_number=2,
        )
        assert options.max_attempts == 3
        assert options.delay == 1.0
        assert options.max_delay == 60.0
        assert options.jitter_type == "none"
        assert options.attempt_number == 2
        # Logger should be a child of the provided logger
        mock_logger.getChild.assert_called_once_with("@teams/retry")


class TestRetry:
    """Test retry functionality."""

    @pytest.mark.asyncio
    async def test_success_on_first_attempt(self):
        """Test that successful operations don't retry."""
        call_count = 0

        async def success_factory():
            nonlocal call_count
            call_count += 1
            return "success"

        result = await retry(success_factory)
        assert result == "success"
        assert call_count == 1

    @pytest.mark.asyncio
    async def test_attempt_number_increments_correctly(self):
        """Test that attempt numbers increment correctly in log messages."""
        call_count = 0
        mock_logger = MagicMock()
        logged_messages = []

        def capture_debug(msg):
            logged_messages.append(msg)

        mock_child_logger = MagicMock()
        mock_child_logger.debug.side_effect = capture_debug
        mock_logger.getChild.return_value = mock_child_logger

        async def failing_factory():
            nonlocal call_count
            call_count += 1
            if call_count < 3:  # Fail first 2 attempts
                raise ValueError(f"Attempt {call_count} failed")
            return "success"

        options = RetryOptions(max_attempts=3, delay=0.01, jitter_type="none", logger=mock_logger)
        result = await retry(failing_factory, options)

        assert result == "success"
        assert call_count == 3

        # Check that attempt numbers are logged correctly
        delay_messages = [msg for msg in logged_messages if "before retry (attempt" in msg]
        assert len(delay_messages) == 2  # Two retries (attempts 1 and 2 failed)
        assert "(attempt 1)" in delay_messages[0]  # First retry shows attempt 1 (which failed)
        assert "(attempt 2)" in delay_messages[1]  # Second retry shows attempt 2 (which failed)

    @pytest.mark.asyncio
    async def test_exponential_backoff_increases(self):
        """Test that exponential backoff delays increase exponentially."""
        call_count = 0
        mock_logger = MagicMock()
        logged_delays = []

        def capture_debug(msg):
            if "Delaying" in msg and "before retry" in msg:
                # Extract delay value from message like "Delaying 1.00s before retry..."
                delay_str = msg.split("Delaying ")[1].split("s")[0]
                logged_delays.append(float(delay_str))

        mock_child_logger = MagicMock()
        mock_child_logger.debug.side_effect = capture_debug
        mock_logger.getChild.return_value = mock_child_logger

        async def failing_factory():
            nonlocal call_count
            call_count += 1
            if call_count < 4:  # Fail first 3 attempts
                raise ValueError(f"Attempt {call_count} failed")
            return "success"

        # Use no jitter to test pure exponential backoff
        options = RetryOptions(max_attempts=4, delay=1.0, jitter_type="none", logger=mock_logger)
        result = await retry(failing_factory, options)

        assert result == "success"
        assert call_count == 4
        assert len(logged_delays) == 3  # Three retries

        # Check exponential increase: base_delay * (2^(attempt_number - 1))
        # Attempt 1: 1.0 * (2^0) = 1.0
        # Attempt 2: 1.0 * (2^1) = 2.0
        # Attempt 3: 1.0 * (2^2) = 4.0
        assert logged_delays[0] == 1.0  # First retry (after attempt 1 failed)
        assert logged_delays[1] == 2.0  # Second retry (after attempt 2 failed)
        assert logged_delays[2] == 4.0  # Third retry (after attempt 3 failed)

    @pytest.mark.asyncio
    async def test_logger_name_consistent(self):
        """Test that logger name doesn't keep growing with recursive calls."""
        call_count = 0
        mock_logger = MagicMock()
        mock_logger.name = "test_logger"

        # Track how many times getChild is called
        child_call_count = 0

        def track_get_child(child_name):
            nonlocal child_call_count
            child_call_count += 1
            child_mock = MagicMock()
            child_mock.name = f"{mock_logger.name}.{child_name}"
            return child_mock

        mock_logger.getChild.side_effect = track_get_child

        async def failing_factory():
            nonlocal call_count
            call_count += 1
            if call_count < 3:  # Fail first 2 attempts
                raise ValueError(f"Attempt {call_count} failed")
            return "success"

        options = RetryOptions(max_attempts=3, delay=0.01, logger=mock_logger)
        result = await retry(failing_factory, options)

        assert result == "success"
        assert call_count == 3

        # getChild should only be called once (for the initial RetryOptions)
        # Recursive calls should reuse the existing child logger
        assert child_call_count == 1
        mock_logger.getChild.assert_called_once_with("@teams/retry")

    @pytest.mark.asyncio
    async def test_final_attempt_failure_raises(self):
        """Test that final attempt failures are raised."""
        call_count = 0

        async def always_fail():
            nonlocal call_count
            call_count += 1
            raise ValueError(f"Always fails - attempt {call_count}")

        options = RetryOptions(max_attempts=2, delay=0.01, jitter_type="none")

        with pytest.raises(ValueError, match="Always fails - attempt 2"):
            await retry(always_fail, options)

        assert call_count == 2  # Should attempt exactly max_attempts times



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""
# pyright: basic

import asyncio
from typing import Optional
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from microsoft.teams.api import (
    Account,
    ConversationAccount,
    InvokeActivity,
    MessageActivity,
    TokenCredentials,
    TokenProtocol,
    TypingActivity,
)
from microsoft.teams.apps import ActivityContext, ActivityEvent, App, AppOptions


class FakeToken(TokenProtocol):
    """Fake token for testing."""

    @property
    def app_id(self) -> str:
        return "test-app-id"

    @property
    def app_display_name(self) -> Optional[str]:
        return "Test App"

    @property
    def tenant_id(self) -> Optional[str]:
        return "test-tenant-id"

    @property
    def service_url(self) -> str:
        return "https://test.service.url"

    @property
    def from_(self) -> str:
        return "azure"

    @property
    def from_id(self) -> str:
        return "test-from-id"

    @property
    def expiration(self) -> Optional[int]:
        return None

    def is_expired(self, buffer_ms: int = 5 * 60 * 1000) -> bool:
        return False

    def __str__(self) -> str:
        return "FakeToken"


class TestApp:
    """Test cases for App class public interface."""

    @pytest.fixture
    def mock_logger(self):
        """Create a mock logger."""
        return MagicMock()

    @pytest.fixture
    def mock_storage(self):
        """Create a mock storage."""
        return MagicMock()

    @pytest.fixture
    def mock_activity_handler(self):
        """Create a mock activity handler."""

        async def handler(ctx) -> None:
            pass

        return handler

    @pytest.fixture(scope="function")
    def basic_options(self, mock_logger, mock_storage):
        """Create basic app options."""
        return AppOptions(
            logger=mock_logger,
            storage=mock_storage,
            client_id="test-client-id",
            client_secret="test-secret",
        )

    def _mock_http_plugin(self, app: App) -> App:
        """Helper to mock the HTTP plugin's create_stream and close methods."""
        mock_stream = MagicMock()
        mock_stream.events = MagicMock()
        mock_stream.events.on = MagicMock()
        mock_stream.close = AsyncMock()
        app.http.create_stream = MagicMock(return_value=mock_stream)
        return app

    @pytest.fixture(scope="function")
    def app_with_options(self, basic_options):
        """Create App with basic options."""
        app = App(**basic_options)
        return self._mock_http_plugin(app)

    @pytest.fixture(scope="function")
    def app_with_activity_handler(self, mock_logger, mock_storage, mock_activity_handler):
        """Create App with activity handler."""
        options = AppOptions(
            logger=mock_logger,
            storage=mock_storage,
            client_id="test-client-id",
            client_secret="test-secret",
            plugins=[],
        )
        app = App(**options)
        app.on_activity(mock_activity_handler)
        return self._mock_http_plugin(app)

    def test_app_starts_successfully(self, basic_options):
        """Test that app can be created and initialized."""
        app = App(**basic_options)

        # Basic functional test - app should be created and not running
        assert not app.is_running
        assert app.port is None

    @pytest.mark.asyncio
    async def test_app_lifecycle_start_stop(self, app_with_options):
        """Test basic app lifecycle: start and stop."""

        # Mock the underlying HTTP server to avoid actual server startup
        async def mock_on_start(event):
            # Simulate the HTTP plugin calling the ready callback
            if app_with_options.http.on_ready_callback:
                await app_with_options.http.on_ready_callback()

        with patch.object(app_with_options.http, "on_start", new_callable=AsyncMock, side_effect=mock_on_start):
            # Test start
            start_task = asyncio.create_task(app_with_options.start(3978))
            await asyncio.sleep(0.1)

            # App should be running and have correct port
            assert app_with_options.is_running
            assert app_with_options.port == 3978

            start_task.cancel()
            try:
                await start_task
            except asyncio.CancelledError:
                pass

        # Test stop
        app_with_options._running = True

        async def mock_on_stop():
            if app_with_options.http.on_stopped_callback:
                await app_with_options.http.on_stopped_callback()

        with patch.object(app_with_options.http, "on_stop", new_callable=AsyncMock, side_effect=mock_on_stop):
            await app_with_options.stop()
            assert not app_with_options.is_running

    # Event Testing - Focus on functional behavior

    @pytest.mark.asyncio
    async def test_activity_event_emission(self, app_with_activity_handler: App) -> None:
        """Test that activity events are emitted correctly."""
        activity_events = []
        event_received = asyncio.Event()

        @app_with_activity_handler.event
        async def handle_activity(event: ActivityEvent) -> None:
            activity_events.append(event)
            event_received.set()

        from_account = Account(id="bot-123", name="Test Bot", role="bot")
        recipient = Account(id="user-456", name="Test User", role="user")
        conversation = ConversationAccount(id="conv-789", conversation_type="personal")

        activity = MessageActivity(
            type="message",
            id="test-activity-id",
            text="Hello, world!",
            from_=from_account,
            recipient=recipient,
            conversation=conversation,
            channel_id="msteams",
        )

        plugin = app_with_activity_handler.http

        await app_with_activity_handler.event_manager.on_activity(
            ActivityEvent(activity=activity, sender=plugin, token=FakeToken()), plugins=[plugin]
        )

        # Wait for the async event handler to complete
        await asyncio.wait_for(event_received.wait(), timeout=1.0)

        # Verify event was emitted
        assert len(activity_events) == 1
        assert isinstance(activity_events[0], ActivityEvent)
        # The event contains the parsed output model, not the input model
        assert activity_events[0].activity.id == activity.id
        assert activity_events[0].activity.type == activity.type
        # Check text only if it's a MessageActivity
        if hasattr(activity_events[0].activity, "text"):
            assert activity_events[0].activity.text == activity.text

    @pytest.mark.asyncio
    async def test_multiple_event_handlers(self, app_with_options: App) -> None:
        """Test that multiple handlers can listen to the same event."""
        activity_events_1 = []
        activity_events_2 = []
        both_received = asyncio.Event()
        received_count = 0

        @app_with_options.event
        async def handle_activity_1(event: ActivityEvent) -> None:
            nonlocal received_count
            activity_events_1.append(event)
            received_count += 1
            if received_count == 2:
                both_received.set()

        @app_with_options.event
        async def handle_activity_2(event: ActivityEvent) -> None:
            nonlocal received_count
            activity_events_2.append(event)
            received_count += 1
            if received_count == 2:
                both_received.set()

        from_account = Account(id="bot-123", name="Test Bot", role="bot")
        recipient = Account(id="user-456", name="Test User", role="user")
        conversation = ConversationAccount(id="conv-789", conversation_type="personal")

        activity = MessageActivity(
            type="message",
            id="test-activity-id",
            text="Hello, world!",
            from_=from_account,
            recipient=recipient,
            conversation=conversation,
            channel_id="msteams",
        )

        await app_with_options.event_manager.on_activity(
            ActivityEvent(activity=activity, sender=app_with_options.http, token=FakeToken()),
            plugins=[app_with_options.http],
        )

        # Wait for both async event handlers to complete
        await asyncio.wait_for(both_received.wait(), timeout=1.0)

        # Both handlers should have received the event
        assert len(activity_events_1) == 1
        assert len(activity_events_2) == 1
        assert activity_events_1[0].activity == activity
        assert activity_events_2[0].activity == activity

    # Generated Handler Tests

    def test_generated_handler_registration(self, app_with_options: App) -> None:
        """Test that generated handlers register correctly in the router."""

        @app_with_options.on_message
        async def handle_message(ctx: ActivityContext[MessageActivity]) -> None:
            assert ctx.activity.type == "message"

        from_account = Account(id="bot-123", name="Test Bot", role="bot")
        recipient = Account(id="user-456", name="Test User", role="user")
        conversation = ConversationAccount(id="conv-789", conversation_type="personal")

        message_activity = MessageActivity(
            id="test-activity-id",
            type="message",
            text="Hello from generated handler!",
            from_=from_account,
            recipient=recipient,
            conversation=conversation,
            channel_id="msteams",
        )

        # Verify handler was registered
        message_handlers = app_with_options.router.select_handlers(message_activity)
        assert len(message_handlers) == 1
        assert message_handlers[0] == handle_message

    def test_multiple_handlers_same_type(self, app_with_options: App) -> None:
        """Test that multiple handlers can be registered for the same activity type."""

        @app_with_options.on_message
        async def handle_message_1(ctx: ActivityContext[MessageActivity]) -> None:
            pass

        @app_with_options.on_message
        async def handle_message_2(ctx: ActivityContext[MessageActivity]) -> None:
            pass

        from_account = Account(id="bot-123", name="Test Bot", role="bot")
        recipient = Account(id="user-456", name="Test User", role="user")
        conversation = ConversationAccount(id="conv-789", conversation_type="personal")

        message_activity = MessageActivity(
            id="test-activity-id",
            type="message",
            text="Hello from generated handler!",
            from_=from_account,
            recipient=recipient,
            conversation=conversation,
            channel_id="msteams",
        )

        # Verify both handlers were registered
        message_handlers = app_with_options.router.select_handlers(message_activity)
        assert len(message_handlers) == 2
        assert handle_message_1 in message_handlers
        assert handle_message_2 in message_handlers

    def test_different_activity_types_separate_routes(self, app_with_options: App) -> None:
        """Test that different activity types are routed separately."""

        @app_with_options.on_message
        async def handle_message(ctx: ActivityContext[MessageActivity]) -> None:
            pass

        @app_with_options.on_typing
        async def handle_typing(ctx: ActivityContext[TypingActivity]) -> None:
            pass

        from_account = Account(id="bot-123", name="Test Bot", role="bot")
        recipient = Account(id="user-456", name="Test User", role="user")
        conversation = ConversationAccount(id="conv-789", conversation_type="personal")

        message_activity = MessageActivity(
            id="test-activity-id",
            type="message",
            text="Hello from generated handler!",
            from_=from_account,
            recipient=recipient,
            conversation=conversation,
            channel_id="msteams",
        )

        typing_activity = TypingActivity(
            id="test-typing-id",
            type="typing",
            from_=from_account,
            recipient=recipient,
            conversation=conversation,
            channel_id="msteams",
        )

        # Verify handlers are in separate routes
        message_handlers = app_with_options.router.select_handlers(message_activity)
        typing_handlers = app_with_options.router.select_handlers(typing_activity)

        assert len(message_handlers) == 1
        assert len(typing_handlers) == 1
        assert message_handlers[0] == handle_message
        assert typing_handlers[0] == handle_typing

    @pytest.mark.asyncio
    async def test_generated_handler_execution(self, app_with_options: App) -> None:
        """Test that generated handlers are executed correctly."""
        handler_data = {}

        @app_with_options.on_message
        async def handle_message(ctx: ActivityContext[MessageActivity]) -> None:
            handler_data["called"] = True
            handler_data["activity_text"] = ctx.activity.text

        from_account = Account(id="bot-123", name="Test Bot", role="bot")
        recipient = Account(id="user-456", name="Test User", role="user")
        conversation = ConversationAccount(id="conv-789", conversation_type="personal")

        activity = MessageActivity(
            id="test-activity-id",
            type="message",
            text="Hello from generated handler!",
            from_=from_account,
            recipient=recipient,
            conversation=conversation,
        )

        # Mock the HTTP plugin response method
        app_with_options.http.on_activity_response = AsyncMock()

        await app_with_options.event_manager.on_activity(
            ActivityEvent(activity=activity, sender=app_with_options.http, token=FakeToken()),
            plugins=[app_with_options.http],
        )

        # Verify handler was called and executed
        assert handler_data["called"] is True
        assert handler_data["activity_text"] == "Hello from generated handler!"

    def test_runtime_type_validation(self, app_with_options: App) -> None:
        """Test that runtime type validation catches incorrect type annotations."""
        with pytest.raises(TypeError) as exc_info:

            @app_with_options.on_message  # type: ignore
            async def handle_wrong_type(ctx: ActivityContext[InvokeActivity]) -> None:  # Wrong type!
                pass

        # Verify the error message mentions the type mismatch
        error_msg = str(exc_info.value)
        assert "InvokeActivity" in error_msg
        assert "MessageActivity" in error_msg
        assert "on_message" in error_msg

    def test_on_message_pattern_string_match(self, app_with_options: App) -> None:
        """Test on_message_pattern with string pattern matching."""

        @app_with_options.on_message_pattern("hello world")
        async def handle_hello(ctx: ActivityContext[MessageActivity]) -> None:
            pass

        from_account = Account(id="bot-123", name="Test Bot", role="bot")
        recipient = Account(id="user-456", name="Test User", role="user")
        conversation = ConversationAccount(id="conv-789", conversation_type="personal")

        # Test matching message
        matching_activity = MessageActivity(
            id="test-activity-id",
            type="message",
            text="hello world",
            from_=from_account,
            recipient=recipient,
            conversation=conversation,
            channel_id="msteams",
        )

        # Test non-matching message
        non_matching_activity = MessageActivity(
            id="test-activity-id-2",
            type="message",
            text="goodbye world",
            from_=from_account,
            recipient=recipient,
            conversation=conversation,
            channel_id="msteams",
        )

        # Verify handler was registered and can match
        handlers = app_with_options.router.select_handlers(matching_activity)
        assert len(handlers) == 1
        assert handlers[0] == handle_hello

        # Verify non-matching activity doesn't match
        non_matching_handlers = app_with_options.router.select_handlers(non_matching_activity)
        assert len(non_matching_handlers) == 0

    def test_on_message_pattern_regex_match(self, app_with_options: App) -> None:
        """Test on_message_pattern with regex pattern matching."""
        import re

        @app_with_options.on_message_pattern(re.compile(r"hello \w+"))
        async def handle_hello_pattern(ctx: ActivityContext[MessageActivity]) -> None:
            pass

        from_account = Account(id="bot-123", name="Test Bot", role="bot")
        recipient = Account(id="user-456", name="Test User", role="user")
        conversation = ConversationAccount(id="conv-789", conversation_type="personal")

        # Test matching message
        matching_activity = MessageActivity(
            id="test-activity-id",
            type="message",
            text="hello world",
            from_=from_account,
            recipient=recipient,
            conversation=conversation,
            channel_id="msteams",
        )

        # Test non-matching message
        non_matching_activity = MessageActivity(
            id="test-activity-id-2",
            type="message",
            text="hello",  # Missing word after hello
            from_=from_account,
            recipient=recipient,
            conversation=conversation,
            channel_id="msteams",
        )

        # Verify handler was registered and can match
        handlers = app_with_options.router.select_handlers(matching_activity)
        assert len(handlers) == 1
        assert handlers[0] == handle_hello_pattern

        # Verify non-matching activity doesn't match
        non_matching_handlers = app_with_options.router.select_handlers(non_matching_activity)
        assert len(non_matching_handlers) == 0

    @pytest.mark.asyncio
    async def test_app_with_callable_token(self):
        """Test that app initializes with callable token."""
        token_called = False

        def get_token(scope, tenant_id=None):
            nonlocal token_called
            token_called = True
            return "test.jwt.token"

        options = AppOptions(client_id="test-client-123", token=get_token)

        # Mock environment variables to ensure they don't interfere
        with patch.dict("os.environ", {"CLIENT_ID": "", "CLIENT_SECRET": "", "TENANT_ID": ""}, clear=False):
            app = App(**options)

            assert app.credentials is not None
            assert type(app.credentials) is TokenCredentials
            assert app.credentials.client_id == "test-client-123"
            assert callable(app.credentials.token)

            res = await app.api.bots.token.get(app.credentials)
            assert token_called is True
            assert res.access_token == "test.jwt.token"

    def test_middleware_registration(self, app_with_options: App) -> None:
        """Test that middleware is registered correctly using app.use()."""

        async def logging_middleware(ctx: ActivityContext) -> None:
            await ctx.next()

        app_with_options.use(logging_middleware)

        from_account = Account(id="bot-123", name="Test Bot", role="bot")
        recipient = Account(id="user-456", name="Test User", role="user")
        conversation = ConversationAccount(id="conv-789", conversation_type="personal")

        message_activity = MessageActivity(
            id="test-activity-id",
            type="message",
            text="hello world",
            from_=from_account,
            recipient=recipient,
            conversation=conversation,
            channel_id="msteams",
        )

        handlers = app_with_options.router.select_handlers(message_activity)
        assert len(handlers) == 1
        assert handlers[0] == logging_middleware

    @pytest.mark.asyncio
    async def test_func_decorator_registration(self, app_with_options: App):
        """Simple test that @app.func registers a function."""
        app_with_options.http.post = MagicMock()

        # Dummy request to simulate a call
        async def dummy_func(ctx):
            return "called"

        decorated = app_with_options.func(dummy_func)
        assert decorated == dummy_func

        # Extract the endpoint path it was registered to
        endpoint_path = app_with_options.http.post.call_args[0][0]
        assert endpoint_path == f"/api/functions/{dummy_func.__name__.replace('_', '-')}"

    def test_user_agent_format(self, app_with_options: App):
        """Test that USER_AGENT follows the expected format teams.py[apps]/{version}."""
        import importlib.metadata

        version = importlib.metadata.version("microsoft-teams-apps")
        expected_user_agent = f"teams.py[apps]/{version}"

        # Verify the http_client has the correct User-Agent header
        assert "User-Agent" in app_with_options.http_client._options.headers
        assert app_with_options.http_client._options.headers["User-Agent"] == expected_user_agent



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass
from logging import Logger
from typing import Generic, Optional, TypeVar

from microsoft.teams.api import (
    Account,
    ActivityParams,
    ApiClient,
    ConversationAccount,
    ConversationReference,
    CreateConversationParams,
    MessageActivityInput,
    SentActivity,
)
from microsoft.teams.cards import AdaptiveCard

from ..http_plugin import HttpPlugin
from .client_context import ClientContext

T = TypeVar("T")


@dataclass(kw_only=True)
class FunctionContext(ClientContext, Generic[T]):
    """
    Context provided to a remote function execution in a Teams app.
    """

    id: Optional[str] = None
    """The ID of the app."""

    name: Optional[str] = None
    """The name of the app."""

    api: ApiClient
    """The API client instance for conversation client."""

    http: HttpPlugin
    """The HTTP plugin instance for sending messages."""

    log: Logger
    """The app logger instance."""

    data: T
    """The function payload."""

    async def send(self, activity: str | ActivityParams | AdaptiveCard) -> Optional[SentActivity]:
        """
        Send an activity to the current conversation.

        Returns None if the conversation ID cannot be determined.
        """
        if self.id is None or self.name is None:
            raise ValueError("app not started")

        conversation_id = await self._resolve_conversation_id(activity)

        if not conversation_id:
            self.log.warning("Cannot send activity: conversation ID could not be resolved")
            return None

        conversation_ref = ConversationReference(
            channel_id="msteams",
            service_url=self.api.service_url,
            bot=Account(id=self.id, name=self.name, role="bot"),
            conversation=ConversationAccount(id=conversation_id, conversation_type="personal"),
        )

        if isinstance(activity, str):
            activity = MessageActivityInput(text=activity)
        elif isinstance(activity, AdaptiveCard):
            activity = MessageActivityInput().add_card(activity)
        else:
            activity = activity

        return await self.http.send(activity, conversation_ref)

    async def _resolve_conversation_id(self, activity: str | ActivityParams | AdaptiveCard) -> Optional[str]:
        """Resolve or create a conversation ID for the current user/context.

        Args:
            activity: The activity to be sent, used to extract conversation info if needed.

        Returns:
            The resolved conversation ID, or None if it could not be determined or created.
        """
        if self._resolved_conversation_id:
            return self._resolved_conversation_id

        self._resolved_conversation_id = self.chat_id or self.channel_id

        # Extract from Activity if available
        if not self._resolved_conversation_id:
            self._resolved_conversation_id = getattr(getattr(activity, "conversation", None), "id", None)

        # Validate that both the bot and user are members of the conversation.
        if self._resolved_conversation_id:
            try:
                member = await self.api.conversations.members_client.get_by_id(
                    self._resolved_conversation_id, self.user_id
                )
                if not member:
                    self.log.warning(
                        f"User {self.user_id} is not a member of conversation {self._resolved_conversation_id}"
                    )
                    self._resolved_conversation_id = None
            except Exception as e:
                self.log.error(f"Failed to get conversation member: {e}")
                self._resolved_conversation_id = None

        else:
            """ Conversation ID can be missing if the app is running in a personal scope. In this case,
            create a conversation between the bot and the user. This will either create a new conversation
            or return a pre-existing one."""
            try:
                conversation_params = CreateConversationParams(
                    bot=Account(id=self.id, name=self.name, role="bot"),  # type: ignore
                    members=[Account(id=self.user_id, role="user", name=self.user_name)],
                    tenant_id=self.tenant_id,
                    is_group=False,
                )
                conversation = await self.api.conversations.create(conversation_params)
                self._resolved_conversation_id = conversation.id
            except Exception as e:
                self.log.error(f"Failed to create conversation: {e}")
                self._resolved_conversation_id = None

        return self._resolved_conversation_id



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""
# pyright: basic

from unittest.mock import AsyncMock, MagicMock

import pytest
from microsoft.teams.api import Activity, ActivityBase, ConversationReference
from microsoft.teams.apps import ActivityContext, Sender
from microsoft.teams.apps.app_events import EventManager
from microsoft.teams.apps.app_process import ActivityProcessor
from microsoft.teams.apps.routing.router import ActivityHandler, ActivityRouter
from microsoft.teams.apps.token_manager import TokenManager
from microsoft.teams.common import Client, ConsoleLogger, LocalStorage


class TestActivityProcessor:
    @pytest.fixture
    def mock_logger(self):
        return MagicMock(spec=ConsoleLogger)

    @pytest.fixture
    def mock_http_client(self):
        return MagicMock(spec=Client)

    @pytest.fixture
    def activity_processor(self, mock_logger, mock_http_client):
        """Create an ActivityProcessor instance."""
        mock_storage = MagicMock(spec=LocalStorage)
        mock_activity_router = MagicMock(spec=ActivityRouter)
        mock_token_manager = MagicMock(spec=TokenManager)
        return ActivityProcessor(
            mock_activity_router,
            mock_logger,
            "id",
            mock_storage,
            "default_connection",
            mock_http_client,
            mock_token_manager,
        )

    @pytest.mark.asyncio
    async def test_execute_middleware_chain_with_no_handlers(self, activity_processor):
        """Test the process_activity method with no handlers."""
        context = MagicMock(spec=ActivityContext)
        activity_processor.event_manager = MagicMock(spec=EventManager)

        response = await activity_processor.execute_middleware_chain(context, [])
        assert response is None

    @pytest.mark.asyncio
    async def test_execute_middleware_chain_with_two_handlers(self, activity_processor, mock_http_client, mock_logger):
        """Test the execute_middleware_chain method with two handlers."""
        context = ActivityContext(
            activity=MagicMock(spec=ActivityBase),
            app_id="app_id",
            logger=mock_logger,
            storage=MagicMock(spec=LocalStorage),
            api=mock_http_client,
            user_token=None,
            conversation_ref=MagicMock(spec=ConversationReference),
            is_signed_in=True,
            connection_name="default_connection",
            sender=MagicMock(spec=Sender),
            app_token=None,
        )

        handler_one = AsyncMock(spec=ActivityHandler)

        async def handler_one_side_effect(ctx: ActivityContext[Activity]) -> str:
            await ctx.next()
            return "handler_one"

        handler_one.side_effect = handler_one_side_effect

        handler_two = AsyncMock(spec=ActivityHandler)

        async def handler_two_side_effect(ctx: ActivityContext[Activity]) -> str:
            await ctx.next()
            return "handler_two"

        handler_two.side_effect = handler_two_side_effect
        handlers = [handler_one, handler_two]

        response = await activity_processor.execute_middleware_chain(context, handlers)
        handler_one.assert_called_once_with(context)
        handler_two.assert_called_once_with(context)
        assert response == "handler_one"



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from dataclasses import dataclass
from typing import Optional


@dataclass
class ClientContext:
    """
    Runtime information about the client and session in Microsoft Teams.
    """

    app_session_id: str
    """Unique ID for the current session for use in correlating telemetry data.
    A session corresponds to the lifecycle of an app.
    A new session begins upon the creation of a webview (on Teams mobile) or iframe (in Teams desktop) hosting the app,
    and ends when it is destroyed."""

    tenant_id: str
    """The Microsoft Entra tenant ID of the current user, extracted from request auth token."""

    user_id: str
    """The Microsoft Entra object ID of the current user, extracted from the request auth token."""

    user_name: str
    """The name of the current user, extracted from the request auth token."""

    page_id: str
    """The developer-defined unique ID for the page this content points to."""

    auth_token: str
    """The MSAL Entra token."""

    app_id: Optional[str] = None
    """This ID is the unique identifier assigned to the app after deployment and is critical for ensuring the correct
    app instance is recognized across hosts."""

    team_id: Optional[str] = None
    """The Microsoft Teams ID for the team with which the content is associated."""

    message_id: Optional[str] = None
    """The ID of the parent message from which this task module was launched.
    This is only available in task modules launched from bot cards."""

    channel_id: Optional[str] = None
    """The Microsoft Teams ID for the channel with which the content is associated."""

    chat_id: Optional[str] = None
    """The Microsoft Teams ID for the chat with which the content is associated."""

    meeting_id: Optional[str] = None
    """Meeting ID used by tab when running in meeting context."""

    sub_page_id: Optional[str] = None
    """The developer-defined unique ID for the sub-page this content points to.
    This field should be used to restore to a specific state within a page,
    such as scrolling to or activating a specific piece of content."""

    _resolved_conversation_id: Optional[str] = None
    """The Microsoft Teams ID for the conversation with which the content is associated."""

    @property
    def conversation_id(self) -> Optional[str]:
        """The Microsoft Teams ID for the conversation with which the content is associated."""
        return self._resolved_conversation_id
